#include "mcmc.h"
#include "DispMessage.h"

using namespace System;
using namespace PoisonExpReg;

namespace mcmc_spc {

	
	double sqr( double x ){
		return x * x;
	}

	double scale_sgm( double acpt_r, int & ck ){
		double v = 1.0;
		ck = 0;
	
		if (acpt_r > 0.5){
			v = 1.0 + 9.0 * (acpt_r - 0.5) / 0.5;
			ck = 1;
		} else if (acpt_r < 0.3){
			v = 1.0 / (1.0 + 9.0 * (0.3 - acpt_r) / 0.3);
			ck = 1;
		}

		return v;
	}

	double myLog( double v ){
		double tv;
		if (v > 1.0E-16){
			tv = Math::Log( v );
		}else{
			tv = Math::Log( 1.0E-16 );
		}
		return tv;
	}


	double myExp( double v ){
		double tv = -1.0;
		if (Math::Abs(v) < 709.0){
			tv = Math::Exp( v );
		}else{
			if(v > 0.0){
				throw gcnew myexcept( "Too large argument Error !", 1 ); 
			}else{
				throw gcnew myexcept( "Too small argument Error ", -1 ); 
			}
		}

		return tv;
	}
	
	void MCMC::Init( int vNSimu, String ^ vmcmcID, RNP2to191 * vrn, int vn, int vp, double * vFrc, double ** vX, double * vsgmB, double * initB ){
		NSimu = vNSimu;
		mcmcID = vmcmcID;
		rn = vrn;
		n = vn;
		p = vp;
		
		Frc = new double[n];
		for (int i = 0; i < n; i++)
			Frc[i] = vFrc[i];
		LogFrcFac = new double[n];
		for (int i = 0; i < n; i++){
			double v = 0.0;
			for (int j = 1; j <= Frc[i]; j++)
				v += Math::Log(j);
			LogFrcFac[i] = v;
		}
		X = new double*[n];
		for (int i = 0; i < n; i++){
			X[i] = new double[p];
			for (int j = 0; j < p; j++)
				X[i][j] = vX[i][j];
		}
		
		sgmB = new double[p];
		for (int i = 0; i < p; i++)
			sgmB[i] = vsgmB[i];
		paramB = new double*[NSimu+1];
		for (int i = 0; i <= NSimu; i++)
			paramB[i] = new double[p];
		for (int i = 0; i < p; i++)
			paramB[0][i] = initB[i];
		countB = new int[p];
		for (int i = 0; i < p; i++)
			countB[i] = 0;
			
	}

	void MCMC::ClearMe(){
	//	if (rn != nullptr) delete rn;
		if (LogFrcFac != nullptr) delete [] LogFrcFac;
		if (Frc != nullptr) delete [] Frc;
		if (X != nullptr){
			for (int i = 0; i < n; i++)
				delete [] X[i];
			delete [] X;
		}
		if (sgmB != nullptr) delete [] sgmB;
		if (paramB != nullptr){
			for (int i = 0; i <= NSimu; i++)
				delete [] paramB[i];
			delete [] paramB;
		}
		if (countB != nullptr) delete [] countB;
	}


	double MCMC::LogL( double * tempB ){
		double tv = 0.0;
		for (int i = 0; i < n; i++){
			double LogLmbd = 0.0;
			for (int j = 0; j < p; j++){
				LogLmbd += tempB[j] * X[i][j];
			}
			tv += -myExp( LogLmbd ) + Frc[i] * LogLmbd - LogFrcFac[i];
		}
		return tv;
	}


	void MCMC::do_mcmc(){
		DispMessage ^ cf = gcnew DispMessage();
		cf->SetMsg( mcmcID );
		cf->Show();

		for (int isimu = 0; isimu < NSimu; isimu++){
			if((isimu % 1000) == 0){
				cf->ShowMessage( " isimu = " + isimu.ToString() + "/" + NSimu.ToString(),
									"Poisson " );
			}

			double * tempB0 = new double[p], * tempB1 = new double[p];
			for (int i = 0; i < p; i++)
				tempB0[i] = tempB1[i] = paramB[isimu][i];

	//		MessageBox::Show("After tempB0...");

			for (int i = 0; i < p; i++){
				double y = rn->normalMS( tempB0[i], sgmB[i] );
				tempB1[i] = y;
				double a;
				try{
					double num = LogL( tempB1 );
					double den = LogL( tempB0 );
					a = myExp( num - den );
				}
				catch( myexcept ^ e ){
					a = -1.0;
				}
				if (rn->uni() < a){
					countB[i]++;
					tempB0[i] = tempB1[i];
				}else{
					tempB1[i] = tempB0[i];
				}
			}
			for (int i = 0; i < p; i++)
				paramB[isimu+1][i] = tempB1[i];

			delete [] tempB0;
			delete [] tempB1;
		}

		cf->Close();
	}


	
	//
	//								Quick Sort
	//
	//      S. Kawai (1988).  Methods of programming (in Japanese). Tokyo: Iwanami-shoten.
	//		S. Ishihata (1989). Algorithms and Data Structure (in Japanese). Tokyo:Iwanami-shoten
	//
	void qSort( double * d, int Lb, int Ub ){
		if (Lb < Ub){
			int p = (Lb + Ub) / 2;
			double c = d[p];
			int i = Lb;
			int j = Ub;
			do {
				while (d[i] < c) i++;
				while (c < d[j]) j--;
				if (i <= j){
					double v = d[i];
					d[i] = d[j];
					d[j] = v;
					i++;
					j--;
				}
			} while (i <= j);
			qSort(d, Lb, j);
			qSort(d, i, Ub);
		}
		return;
	}

}